###############################################################
# Name:             SongpressFrame.py
# Purpose:     Main frame for Songpress
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2009-01-16
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import subprocess
import os
import os.path
import platform

# import wx.aui as aui
import wx.adv
from wx import xrc

from .Editor import *
from .PreviewCanvas import *
from .Renderer import *
from .FontComboBox import FontComboBox
from .FontFaceDialog import FontFaceDialog
from .MyPreferencesDialog import MyPreferencesDialog
from .HTML import HtmlExporter, TabExporter
from .MyTransposeDialog import *
from .MyNotationDialog import *
from .MyNormalizeDialog import *
from .MyListDialog import MyListDialog
from . import MyUpdateDialog
from .Globals import glb
from .Preferences import Preferences
from . import i18n
from .utils import temp_dir

_ = wx.GetTranslation


if platform.system() == 'Windows':
    import wx.msw


class SongpressFindReplaceDialog(object):
    def __init__(self, owner, replace=False):
        object.__init__(self)
        self.down = True
        self.st = ''
        self.data = wx.FindReplaceData(wx.FR_DOWN)
        self.owner = owner
        self.flags = 0
        self.dialog = wx.FindReplaceDialog(
            owner.frame,
            self.data,
            _("Replace") if replace else _("Find"),
            wx.FR_REPLACEDIALOG if replace else 0
        )
        owner.frame.Bind(wx.EVT_FIND, self.OnFind, self.dialog)
        owner.frame.Bind(wx.EVT_FIND_NEXT, self.OnFind, self.dialog)
        owner.frame.Bind(wx.EVT_FIND_CLOSE, self.OnClose, id=wx.ID_ANY)
        owner.frame.Bind(wx.EVT_FIND_REPLACE, self.OnReplace, self.dialog)
        owner.frame.Bind(wx.EVT_FIND_REPLACE_ALL, self.OnReplaceAll, self.dialog)
        self.dialog.Show()

    def OnFind(self, evt):
        self.st = self.data.GetFindString()
        f = self.data.GetFlags()
        self.down = f & wx.FR_DOWN
        self.whole = f & wx.FR_WHOLEWORD
        self.case = f & wx.FR_MATCHCASE
        self.flags = 0
        if self.whole:
            self.flags |= wx.stc.STC_FIND_WHOLEWORD
        if self.case:
            self.flags |= wx.stc.STC_FIND_MATCHCASE
        self.FindNext()

    def FindNext(self):
        if self.down:
            self.search = self.owner.text.SearchNext
        else:
            self.search = self.owner.text.SearchPrev
        s, e = self.owner.text.GetSelection()
        if self.down:
            self.owner.text.SetSelection(e, e)
            fromStart = s == 0
        else:
            self.owner.text.SetSelection(s, s)
            fromStart = s == self.owner.text.GetLength()
        self.owner.text.SearchAnchor()
        p = self.search(self.flags, self.st)
        if p != -1:
            pass
            self.owner.text.SetSelection(p, p + len(self.st))
        else:
            parent = self.dialog if self.dialog != None else self.owner.frame
            if not fromStart:
                if self.down:
                    where = _("beginning")
                    wherefrom = _("Reached the end")
                    newStart = 0
                else:
                    where = _("end")
                    wherefrom = _("Reached the beginning")
                    newStart = self.owner.text.GetLength()
                d = wx.MessageDialog(
                    parent,
                    _("%s of the song, restarting search from the %s") % (wherefrom, where,),
                    self.owner.appName,
                    wx.OK | wx.CANCEL | wx.ICON_INFORMATION
                )
                res = d.ShowModal()
                if res == wx.ID_OK:
                    self.owner.text.SetSelection(newStart, newStart)
                    self.FindNext()
            else:
                d = wx.MessageDialog(
                    parent,
                    _("The specified text was not found"),
                    self.owner.appName,
                    wx.OK | wx.ICON_INFORMATION
                )
                res = d.ShowModal()

    def OnReplace(self, evt):
        r = self.data.GetReplaceString()
        self.st = self.data.GetFindString()
        if self.owner.text.GetSelectedText().lower() == self.st.lower():
            self.owner.text.ReplaceSelection(r)
            self.FindNext()

    def OnReplaceAll(self, evt):
        self.owner.text.BeginUndoAction()
        s = self.data.GetFindString()
        r = self.data.GetReplaceString()
        f = self.data.GetFlags()
        self.whole = f & wx.FR_WHOLEWORD
        self.case = f & wx.FR_MATCHCASE
        flags = 0
        if self.whole:
            flags |= wx.stc.STC_FIND_WHOLEWORD
        if self.case:
            flags |= wx.stc.STC_FIND_MATCHCASE
        self.owner.text.SetSelection(0, 0)
        c = 0
        p = 0
        while (p := self.owner.text.FindText(p, self.owner.text.GetLength(), s, flags)[0]) != -1:
            self.owner.text.SetTargetStart(p)
            self.owner.text.SetTargetEnd(p + len(s))
            p += self.owner.text.ReplaceTarget(r)
            c += 1

        self.owner.text.EndUndoAction()

        d = wx.MessageDialog(
            self.dialog,
            _("%d text occurrences have been replaced") % (c,),
            self.owner.appName,
            wx.OK | wx.ICON_INFORMATION
        )
        d.ShowModal()

    def OnClose(self, evt):
        self.dialog.Destroy()
        self.dialog = None


if platform.system() == 'Linux':
    # Apparently there is a problem with linux FileOpen dialog box in wxPython:
    # it does not support multiple extensions in a filter.
    _import_formats = [
        (_("All supported files"), ["crd", "cho", "chordpro", "chopro", "tab", "cpm"]),
        #(_("Chordpro files (*.crd)"), ["crd"]),
        #(_("Tab files (*.tab)"), ["tab"]),
        #(_("Chordpro files (*.cho)"), ["cho"]),
        #(_("Chordpro files (*.chordpro)"), ["chordpro"]),
        #(_("Chordpro files (*.chopro)"), ["chopro"]),
        #(_("Chordpro files (*.pro)"), ["pro"]),
    ]
else:
    _import_formats = [
        (_("All supported files"), ["crd", "cho", "chordpro", "chopro", "pro", "tab"]),
        (_("Chordpro files (*.crd, *.cho, *.chordpro, *.chopro, *.pro)"), ["crd", "cho", "chordpro", "chopro", "pro"]),
        (_("Tab files (*.tab)"), ["tab"]),
    ]


class SongpressFrame(SDIMainFrame):
    def __init__(self, res):
        SDIMainFrame.__init__(
            self,
            res,
            'MainFrame',
            'songpress',
            'Skeed',
            _('song'),
            'crd',
            _('Songpress - Il Canzonatore'),
            glb.AddPath('img/songpress.ico'),
            glb.VERSION,
            _("http://www.skeed.it/songpress"),
            (_(u"Copyright (c) 2009-{year} Luca Allulli - Skeed\nLocalization:\n{translations}")).format(
                year=glb.YEAR,
                translations="\n".join([u"- {}: {}".format(glb.languages[x], glb.translators[x]) for x in glb.languages])
            ),
            _("Licensed under the terms and conditions of the GNU General Public License, version 2"),
            _(
                "Special thanks to:\n  * The Pyhton programming language (http://www.python.org)\n  * wxWidgets (http://www.wxwidgets.org)\n  * wxPython (http://www.wxpython.org)\n  * Editra (http://editra.org/) (for the error reporting dialog and... the editor itself!)\n  * python-pptx (for PowerPoint export)"),
            _import_formats,
        )
        self.pref = Preferences()
        self.SetDefaultExtension(self.pref.defaultExtension)
        self.text = Editor(self)
        dt = SDIDropTarget(self)
        self.text.SetDropTarget(dt)
        self.frame.Bind(wx.stc.EVT_STC_UPDATEUI, self.OnUpdateUI, self.text)
        self.text.Bind(wx.EVT_KEY_DOWN, self.OnTextKeyDown, self.text)
        # Other objects
        self.previewCanvas = PreviewCanvas(self.frame, self.pref.format, self.pref.notations, self.pref.decorator)
        self.AddMainPane(self.text)
        self.AddPane(self.previewCanvas.main_panel, aui.AuiPaneInfo().Right().BestSize(240, 400), _('Preview'), 'preview')
        self.previewCanvas.main_panel.Bind(wx.adv.EVT_HYPERLINK, self.OnCopyAsImage, self.previewCanvas.link)
        self.mainToolBar = aui.AuiToolBar(self.frame, wx.ID_ANY, wx.DefaultPosition, agwStyle=aui.AUI_TB_PLAIN_BACKGROUND)
        self.mainToolBar.SetToolBitmapSize(wx.Size(16, 16))
        self.AddTool(self.mainToolBar, 'new', 'img/new.png', _("New"), _("Create a new song"))
        self.AddTool(self.mainToolBar, 'open', 'img/open.png', _("Open"), _("Open an existing song"))
        self.AddTool(self.mainToolBar, 'save', 'img/save.png', _("Save"), _("Save song with the current filename"))
        self.mainToolBar.AddSeparator()
        self.undoTool = self.AddTool(self.mainToolBar, 'undo', 'img/undo.png', _("Undo"), _("Undo last edit"))
        self.redoTool = self.AddTool(self.mainToolBar, 'redo',
            'img/redo.png', _("Redo"), _("Redo previously undone edit"))
        self.redoTool = wx.xrc.XRCID('redo')
        self.mainToolBar.AddSeparator()
        self.cutTool = self.AddTool(self.mainToolBar, 'cut', 'img/cut.png', _("Cut"),
                                                                _("Move selected text in the clipboard"))
        self.copyTool = self.AddTool(self.mainToolBar, 'copy', 'img/copy.png', _("Copy"),
                                                                 _("Copy selected text in the clipboard"))
        self.copyOnlyTextTool = wx.xrc.XRCID('copyOnlyText')
        self.AddTool(self.mainToolBar, 'copyAsImage', 'img/copyAsImage2.png', _("Copy as Image"),
                                     _("Copy the whole FORMATTED song (or selected verses) to the clipboard"))
        self.pasteTool = self.AddTool(self.mainToolBar, 'paste', 'img/paste.png', _("Paste"),
                                                                    _("Read text from the clipboard and place it at the cursor position"))
        self.pasteChordsTool = self.AddTool(self.mainToolBar, 'pasteChords', 'img/pasteChords.png', _("PasteChords"),
                                                                                _("Integrate chords of copied text into current selection"))
        self.mainToolBar.Realize()
        self.mainToolBarPane = self.AddPane(self.mainToolBar, aui.AuiPaneInfo().ToolbarPane().Top().Row(1).Position(1),
                                                                                _('Standard'), 'standard')
        self.formatToolBar = aui.AuiToolBar(self.frame, wx.ID_ANY, agwStyle=aui.AUI_TB_PLAIN_BACKGROUND)
        self.formatToolBar.SetExtraStyle(aui.AUI_TB_PLAIN_BACKGROUND)
        self.fontChooser = FontComboBox(self.formatToolBar, -1, self.pref.format.face)
        self.formatToolBar.AddControl(self.fontChooser)
        self.frame.Bind(wx.EVT_COMBOBOX, self.OnFontSelected, self.fontChooser)
        wx.UpdateUIEvent.SetUpdateInterval(500)
        self.frame.Bind(wx.EVT_UPDATE_UI, self.OnIdle, self.frame)
        self.frame.Bind(wx.EVT_TEXT_CUT, self.OnTextCutCopy, self.text)
        self.frame.Bind(wx.EVT_TEXT_COPY, self.OnTextCutCopy, self.text)
        self.fontChooser.Bind(wx.EVT_TEXT_ENTER, self.OnFontSelected, self.fontChooser)
        self.fontChooser.Bind(wx.EVT_KILL_FOCUS, self.OnFontSelected, self.fontChooser)
        self.AddTool(self.formatToolBar, 'title', 'img/title.png', _("Insert title"),
                                 _("Insert a command to display song title"))
        self.AddTool(self.formatToolBar, 'chord', 'img/chord.png', _("Insert chord"),
                                 _("Insert square brackets that will host a chord"))
        self.AddTool(self.formatToolBar, 'chorus', 'img/chorus.png', _("Insert chorus"),
                                 _("Insert a couple of commands that will contain chorus"))
        self.AddTool(
            self.formatToolBar,
            'verseWithCustomLabelOrWithoutLabel',
            'img/verse.png',
            _("Insert verse with custom label or without label"),
            _("Insert a commands that will display a verse with a custom label"),
        )
        labelVersesTool = self.formatToolBar.AddToggleTool(  # AddToggleTool (agw) or AddTool
            wx.xrc.XRCID('labelVerses'),
            wx.Bitmap(wx.Image(glb.AddPath("img/labelVerses.png"))),
            wx.NullBitmap,
            True,
            None,
            _("Show verse labels"),
            _("Show or hide verse and chorus labels"),
        )
        self.labelVersesToolId = labelVersesTool.GetId()
        showChordsIcon = wx.StaticBitmap(self.formatToolBar, -1, wx.Bitmap(wx.Image(glb.AddPath('img/showChords.png'))))
        self.formatToolBar.AddControl(showChordsIcon)
        self.showChordsChooser = wx.Slider(self.formatToolBar, -1, 0, 0, 2, wx.DefaultPosition, (100, -1),
                                                                             wx.SL_AUTOTICKS | wx.SL_HORIZONTAL)
        tt1 = wx.ToolTip(_("Hide or show chords in formatted song"))
        tt2 = wx.ToolTip(_("Hide or show chords in formatted song"))
        self.showChordsChooser.SetToolTip(tt1)
        showChordsIcon.SetToolTip(tt2)
        self.frame.Bind(wx.EVT_SCROLL, self.OnFontSelected, self.showChordsChooser)
        self.formatToolBar.AddControl(
            self.showChordsChooser,
            "pippo"
        )
        self.formatToolBar.Realize()
        self.formatToolBarPane = self.AddPane(self.formatToolBar, aui.AuiPaneInfo().ToolbarPane().Top().Row(1).Position(2),
                                                                                    _('Format'), 'format')
        self.BindMyMenu()
        self.frame.Bind(EVT_TEXT_CHANGED, self.OnTextChanged)
        self.exportMenuId = xrc.XRCID('export')
        self.exportToClipboardAsAVectorImage = xrc.XRCID('exportToClipboardAsAVectorImage')
        self.exportAsEmfMenuId = xrc.XRCID('exportAsEmf')
        self.cutMenuId = xrc.XRCID('cut')
        self.copyMenuId = xrc.XRCID('copy')
        self.copyAsImageMenuId = xrc.XRCID('copyAsImage')
        self.pasteMenuId = xrc.XRCID('paste')
        self.pasteChordsMenuId = xrc.XRCID('pasteChords')
        self.removeChordsMenuId = xrc.XRCID('removeChords')
        self.labelVersesMenuId = xrc.XRCID('labelVerses')
        self.noChordsMenuId = xrc.XRCID('noChords')
        self.oneVerseForEachChordPatternMenuId = xrc.XRCID('oneVerseForEachChordPattern')
        self.wholeSongMenuId = xrc.XRCID('wholeSong')
        if platform.system() != 'Windows':
            self.menuBar.GetMenu(0).FindItemById(self.exportMenuId).GetSubMenu().Delete(self.exportAsEmfMenuId)
        self.findReplaceDialog = None
        self.CheckLabelVerses()
        self.SetFont()
        self.text.SetFont(self.pref.editorFace, self.pref.editorSize)
        self.FinalizePaneInitialization()
        # Reassign caption value to override caption saved in preferences (it could be another language)
        self._mgr.GetPane('preview').caption = _('Preview')
        self._mgr.GetPane('standard').caption = _('Standard')
        self._mgr.GetPane('format').caption = _('Format')
        if 'firstTimeEasyKey' in self.pref.notices:
            msg = _(
                "You are not a skilled guitarist? Songpress can help you: when you open a song, it can detect if chords are difficult. If this is the case, Songpress will alert you, and offer to transpose your song to the easiest key, automatically.\n\nDo you want to turn this option on?")
            d = wx.MessageDialog(self.frame, msg, _("Songpress"), wx.YES_NO | wx.ICON_QUESTION)
            if d.ShowModal() == wx.ID_YES:
                self.pref.autoAdjustEasyKey = True
                msg = _(
                    "Please take a minute to set up your skill as a guitarist. For each group of chords, tell Songpress how much you like them.")
                d = wx.MessageDialog(self.frame, msg, _("Songpress"), wx.OK)
                d.ShowModal()
                f = MyPreferencesDialog(self.frame, self.pref, easyChords)
                f.notebook.SetSelection(1)
                if f.ShowModal() == wx.ID_OK:
                    self.text.SetFont(self.pref.editorFace, int(self.pref.editorSize))
                    self.SetDefaultExtension(self.pref.defaultExtension)
        MyUpdateDialog.check_and_update(self.frame, self.pref)

    def OnClose(self, evt):
        self.config.Flush()
        super().OnClose(evt)

    def BindMyMenu(self):
        """Bind a menu item, by xrc name, to a handler"""

        def Bind(handler, xrcname):
            self.Bind(wx.EVT_MENU, handler, xrcname)

        Bind(self.OnCopyAsImage, 'exportToClipboardAsAVectorImage')
        Bind(self.OnExportAsSvg, 'exportAsSvg')
        Bind(self.OnExportAsEmf, 'exportAsEmf')
        Bind(self.OnExportAsPng, 'exportAsPng')
        Bind(self.OnExportAsHtml, 'exportAsHtml')
        Bind(self.OnExportAsTab, 'exportAsTab')
        Bind(self.OnExportAsPptx, 'exportAsPptx')
        Bind(self.OnUndo, 'undo')
        Bind(self.OnRedo, 'redo')
        Bind(self.OnCut, 'cut')
        Bind(self.OnCopy, 'copy')
        Bind(self.OnCopyAsImage, 'copyAsImage')
        Bind(self.OnCopyOnlyText, 'copyOnlyText')
        Bind(self.OnPaste, 'paste')
        Bind(self.OnPasteChords, 'pasteChords')
        Bind(self.OnFind, 'find')
        Bind(self.OnFindNext, 'findNext')
        Bind(self.OnFindPrevious, 'findPrevious')
        Bind(self.OnReplace, 'replace')
        Bind(self.OnSelectAll, 'selectAll')
        Bind(self.OnSelectNextChord, 'selectNextChord')
        Bind(self.OnSelectPreviousChord, 'selectPreviousChord')
        Bind(self.OnMoveChordRight, 'moveChordRight')
        Bind(self.OnMoveChordLeft, 'moveChordLeft')
        Bind(self.OnRemoveChords, 'removeChords')
        Bind(self.OnIntegrateChords, 'integrateChords')
        Bind(self.OnTitle, 'title')
        Bind(self.OnSubtitle, 'subtitle')
        Bind(self.OnChord, 'chord')
        Bind(self.OnChorus, 'chorus')
        Bind(self.OnVerse, 'verseWithCustomLabelOrWithoutLabel')
        Bind(self.OnComment, 'comment')
        Bind(self.OnFormatFont, 'songFont')
        Bind(self.OnTextFont, 'textFont')
        Bind(self.OnChordFont, 'chordFont')
        Bind(self.OnLabelVerses, 'labelVerses')
        Bind(self.OnChorusLabel, 'chorusLabel')
        Bind(self.OnNoChords, 'noChords')
        Bind(self.OnOneVerseForEachChordPattern, 'oneVerseForEachChordPattern')
        Bind(self.OnWholeSong, 'wholeSong')
        Bind(self.OnTranspose, 'transpose')
        Bind(self.OnSimplifyChords, 'simplifyChords')
        Bind(self.OnChangeChordNotation, 'changeChordNotation')
        Bind(self.OnNormalizeChords, 'cleanupChords')
        Bind(self.OnConvertTabToChordpro, 'convertTabToChordpro')
        Bind(self.OnRemoveSpuriousBlankLines, 'removeSpuriousBlankLines')
        Bind(self.OnOptions, 'options')
        Bind(self.OnGuide, 'guide')
        Bind(self.OnNewsAndUpdates, 'newsAndUpdates')
        Bind(self.OnDonate, 'donate')

    def AddTool(self, toolbar, resource_string, icon_path, label, help):
        tool = wx.xrc.XRCID(resource_string)
        toolbar.AddTool(
            tool,
            label,
            wx.Bitmap(wx.Image(glb.AddPath(icon_path))),
            wx.NullBitmap,
            wx.ITEM_NORMAL,
            label,
            help,
            None
        )
        return tool

    def New(self):
        self.text.AutoChangeMode(True)
        self.text.New()
        self.text.AutoChangeMode(False)
        self.UpdateEverything()

    def Open(self):
        self.text.AutoChangeMode(True)
        self.text.Open()
        self.text.AutoChangeMode(False)
        self.UpdateEverything()
        self.AutoAdjust(0, self.text.GetLength())

    def Save(self):
        self.text.Save()
        self.UpdateEverything()

    def SavePreferences(self):
        self.pref.Save()

    def UpdateUndoRedo(self):
        self.mainToolBar.EnableTool(self.undoTool, self.text.CanUndo())
        self.mainToolBar.EnableTool(self.redoTool, self.text.CanRedo())

    def UpdateCutCopyPaste(self):
        s, e = self.text.GetSelection()
        self.mainToolBar.EnableTool(self.cutTool, s != e)
        self.menuBar.Enable(self.cutMenuId, s != e)
        self.mainToolBar.EnableTool(self.copyTool, s != e)
        self.menuBar.Enable(self.copyOnlyTextTool, s != e)
        self.menuBar.Enable(self.copyMenuId, s != e)
        if platform.system() == 'Windows':
            cp = self.text.CanPaste()
        else:
            # Workaround for weird error in wxGTK
            cp = True
        self.mainToolBar.EnableTool(self.pasteTool, cp)
        self.menuBar.Enable(self.pasteMenuId, cp)
        self.mainToolBar.EnableTool(self.pasteChordsTool, cp)
        self.menuBar.Enable(self.pasteChordsMenuId, cp)
        self.menuBar.Enable(self.removeChordsMenuId, s != e)

    def UpdateEverything(self):
        self.UpdateUndoRedo()
        self.UpdateCutCopyPaste()

    def TextUpdated(self):
        self.previewCanvas.Refresh(self.text.GetText())

    # self.UpdateEverything()

    def DrawOnDC(self, dc):
        """
        Draw song on DC and return the tuple (width, height) of rendered song
        """
        decorator = self.pref.decorator if self.pref.labelVerses else SongDecorator()
        r = Renderer(self.pref.format, decorator, self.pref.notations)
        start, end = self.text.GetSelection()
        if start == end:
            w, h = r.Render(self.text.GetText(), dc)
        else:
            w, h = r.Render(self.text.GetText(), dc, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
        return w, h

    def AskExportFileName(self, type, ext):
        """Ask the filename (without saving); return None if user cancels, the file name ow"""
        leave = False;
        consensus = False;
        while not leave:
            dlg = wx.FileDialog(
                self.frame,
                "Choose a name for the output file",
                "",
                os.path.splitext(self.document)[0],
                "%s files (*.%s)|*.%s|All files (*.*)|*.*" % (type, ext, ext),
                wx.FD_SAVE
            )

            if dlg.ShowModal() == wx.ID_OK:

                fn = dlg.GetPath()
                if os.path.isfile(fn):
                    msg = "File \"%s\" already exists. Do you want to overwrite it?" % (fn,)
                    d = wx.MessageDialog(
                        self.frame,
                        msg,
                        self.appLongName,
                        wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION
                    )
                    res = d.ShowModal()
                    if res == wx.ID_CANCEL:
                        leave = True
                        consensus = False
                    elif res == wx.ID_NO:
                        leave = False
                        consensus = False
                    else:  # wxID_YES
                        leave = True
                        consensus = True
                else:
                    leave = True
                    consensus = True

            else:
                leave = True
                consensus = False

        if consensus:
            return fn
        else:
            return None

    def ComputeRenderedSize(self):
        """
        Compute and return rendered size as tuple (with, height)
        """
        dc = wx.MemoryDC(wx.Bitmap(1, 1))
        w, h = self.DrawOnDC(dc)
        return max(1, w), max(1, h)

    def RenderAsPng(self, scale=1, size=None):
        if size is None:
            size = self.ComputeRenderedSize()
        w, h = size
        b = wx.Bitmap(int(w * scale), int(h * scale))
        dc = wx.MemoryDC(b)
        dc.SetUserScale(scale, scale)
        dc.SetBackground(wx.WHITE_BRUSH)
        dc.Clear()
        self.DrawOnDC(dc)
        return b

    def OnExportAsPng(self, evt):
        n = self.AskExportFileName(_("PNG image"), "png")
        if n is not None:
            b = self.RenderAsPng()
            i = b.ConvertToImage()
            i.SaveFile(n, wx.BITMAP_TYPE_PNG)

    def OnExportAsHtml(self, evt):
        n = self.AskExportFileName(_("HTML file"), "html")
        if n is not None:
            h = HtmlExporter(self.pref.format)
            r = Renderer(self.pref.format, h, self.pref.notations)
            start, end = self.text.GetSelection()
            if start == end:
                r.Render(self.text.GetText(), None)
            else:
                r.Render(self.text.GetText(), None, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
            with open(n, "w", encoding='utf-8') as f:
                f.write(h.getHtml())

    def OnExportAsTab(self, evt):
        n = self.AskExportFileName(_("TAB file"), "tab")
        if n is not None:
            t = TabExporter(self.pref.format)
            r = Renderer(self.pref.format, t, self.pref.notations)
            start, end = self.text.GetSelection()
            if start == end:
                r.Render(self.text.GetText(), None)
            else:
                r.Render(self.text.GetText(), None, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
            with open(n, "w", encoding='utf-8') as f:
                f.write(t.getTab())

    def SaveSvg(self, filename, size=None):
        if size is None:
            size = self.ComputeRenderedSize()
        w, h = size
        dc = wx.SVGFileDC(filename, int(w), int(h))
        self.DrawOnDC(dc)

    def OnExportAsSvg(self, evt):
        n = self.AskExportFileName(_("SVG image"), "svg")
        if n is not None:
            self.SaveSvg(n)

    def OnExportAsEmf(self, evt):
        n = self.AskExportFileName(_("Enhanced Metafile"), "emf")
        if n is not None:
            dc = wx.msw.MetafileDC(n)
            self.DrawOnDC(dc)
            dc.Close()

    def OnExportAsEps(self, evt):
        n = self.AskExportFileName(_("EPS image"), "eps")
        if n is not None:
            pd = wx.PrintData()
            pd.SetPaperId(wx.PAPER_NONE)
            pd.SetPrintMode(wx.PRINT_MODE_FILE)
            pd.SetFilename(n)
            dc = wx.PostScriptDC(pd)
            dc.StartDoc(_("Exporting image as EPS..."))
            self.DrawOnDC(dc)
            dc.EndDoc()

    def OnExportAsPptx(self, evt):
        try:
            from . import songimpress
        except ImportError:
            msg = _("Please install the python-pptx module to use this feature")
            d = wx.MessageDialog(self.frame, msg, "Songpress", wx.OK | wx.ICON_ERROR)
            d.ShowModal()
            return
        text = replaceTitles(self.text.GetTextOrSelection(), '---')
        text = removeChordPro(text).strip()
        if text != '':
            template_rel = os.path.join('templates', 'slides')
            template_paths = [f for f in glb.ListLocalGlobalDir(template_rel) if f[-5:].upper() == '.PPTX']
            template_names = [os.path.split(f)[1][:-5] for f in template_paths]
            mld = MyListDialog(
                self.frame,
                _("Please select a template for your PowerPoint presentation:"),
                _("Export as PowerPoint"),
                template_names,
            )
            if mld.ShowModal() == wx.ID_OK:
                output_file = self.AskExportFileName(_("PPTX presentation"), "pptx")
                if output_file is not None:
                    i = mld.GetSelectedIndex()
                    songimpress.to_presentation(text.splitlines(), output_file, template_paths[i])

    def OnUpdateUI(self, evt):
        self.UpdateEverything()
        evt.Skip()

    def OnUndo(self, evt):
        if self.text.CanUndo():
            self.text.Undo()
            self.UpdateUndoRedo()

    def OnRedo(self, evt):
        if self.text.CanRedo():
            self.text.Redo()
            self.UpdateUndoRedo()

    def OnCut(self, evt):
        self.text.Cut()

    def OnTextCutCopy(self, evt):
        self.UpdateCutCopyPaste()
        evt.Skip()

    def OnTextKeyDown(self, evt):
        # 314: left
        # 316: right
        map = {
            (314, True, True, False): self.MoveChordLeft,
            (316, True, True, False): self.MoveChordRight,
            (ord('D'), False, False, True): self.CopyAsImage,
        }
        tp = (
            evt.GetKeyCode(),
            evt.ShiftDown(),
            evt.AltDown(),
            evt.ControlDown(),
        )
        if (method := map.get(tp)) is not None:
            method()
            evt.Skip(False)
        else:
            evt.Skip()

    def Copy(self):
        self.text.Copy()

    def OnCopy(self, evt):
        self.Copy()

    def OnCopyOnlyText(self, evt):
        self.text.CopyOnlyText()

    def CopyAsImage(self):
        if platform.system() == 'Windows':
            # Windows Metafile
            dc = wx.MetafileDC()
            self.DrawOnDC(dc)
            m = dc.Close()
            m.SetClipboard(dc.MaxX(), dc.MaxY())

        else:
            composite = wx.DataObjectComposite()
            size = self.ComputeRenderedSize()

            # 1. SVG
            with temp_dir() as path:
                svg_obj = wx.CustomDataObject("image/svg+xml")
                fp = os.path.join(path, 'temp.svg')
                self.SaveSvg(fp, size=size)
                with open(fp, 'rb') as f:
                    svg_obj.SetData(f.read())
                composite.Add(svg_obj, preferred=True)

            # 2. PNG
            bmp = self.RenderAsPng(scale=2, size=size)
            png_obj = wx.BitmapDataObject(bmp)
            composite.Add(png_obj)

            # Place on Clipboard
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(composite)
                wx.TheClipboard.Close()

    def OnCopyAsImage(self, evt):
        self.CopyAsImage()

    def OnPaste(self, evt):
        self.text.Paste()

    def OnPasteChords(self, evt):
        self.text.PasteChords()

    def OnFind(self, evt):
        self.findReplaceDialog = SongpressFindReplaceDialog(self)

    def OnFindNext(self, evt):
        if self.findReplaceDialog != None:
            self.findReplaceDialog.down = True
            self.findReplaceDialog.FindNext()

    def OnFindPrevious(self, evt):
        if self.findReplaceDialog != None:
            self.findReplaceDialog.down = False
            self.findReplaceDialog.FindNext()

    def OnReplace(self, evt):
        self.findReplaceDialog = SongpressFindReplaceDialog(self, True)

    def OnSelectAll(self, evt):
        self.text.SelectAll()

    def OnSelectNextChord(self, evt):
        self.text.SelectNextChord()

    def OnSelectPreviousChord(self, evt):
        self.text.SelectPreviousChord()

    def MoveChordRight(self):
        r = self.text.GetChordUnderCursor()
        if r is not None:
            n = self.text.GetLength()
            s, e, c = r
            if e < n:
                self.text.BeginUndoAction()
                e1 = self.text.PositionAfter(e)
                self.text.SetSelection(e, e1)
                l = self.text.GetTextRange(e, e1)
                self.text.ReplaceSelection('')
                self.text.SetSelection(s, s)
                self.text.ReplaceSelection(l)
                s2 = self.text.PositionAfter(self.text.PositionAfter(s))
                self.text.SetSelection(s2, s2)
                self.text.EndUndoAction()

    def OnMoveChordRight(self, evt):
        self.MoveChordRight()

    def MoveChordLeft(self):
        r = self.text.GetChordUnderCursor()
        if r is not None:
            s, e, c = r
            if s > 0:
                self.text.BeginUndoAction()
                s1 = self.text.PositionBefore(s)
                e1 = self.text.PositionBefore(e)
                l = self.text.GetTextRange(s1, s)
                self.text.SetSelection(e, e)
                self.text.ReplaceSelection(l)
                self.text.SetSelection(s1, s)
                self.text.ReplaceSelection('')
                s = self.text.PositionAfter(s1)
                self.text.SetSelection(s, s)
                self.text.EndUndoAction()

    def OnMoveChordLeft(self, evt):
        self.MoveChordLeft()

    def OnRemoveChords(self, evt):
        self.text.RemoveChordsInSelection()

    def OnIntegrateChords(self, evt):
        ln = self.text.GetCurrentLine()
        if ln < self.text.GetLineCount() - 1:
            chords = self.text.GetLine(ln).strip("\r\n")
            text = self.text.GetLine(ln + 1).strip("\r\n")
            chordpro = integrateChords(chords, text)
            self.text.SetSelectionStart(self.text.PositionFromLine(ln))
            self.text.SetSelectionEnd(self.text.GetLineEndPosition(ln + 1))
            self.text.ReplaceSelection(chordpro)

    def OnFontSelected(self, evt):
        font = self.fontChooser.GetValue()
        showChords = self.showChordsChooser.GetValue()
        self.pref.SetFont(font, showChords)
        self.SetFont(True)
        evt.Skip()

    def OnGuide(self, evt):
        wx.LaunchDefaultBrowser(_("http://www.skeed.it/songpress-manual"))

    def OnIdle(self, evt):
        try:
            cp = self.text.CanPaste()
            self.mainToolBar.EnableTool(self.pasteTool, cp)
            self.menuBar.Enable(self.pasteMenuId, cp)
            self.mainToolBar.EnableTool(self.pasteChordsTool, cp)
            self.menuBar.Enable(self.pasteChordsMenuId, cp)
        except Exception:
            # When frame is closed, this method may still be executed, generating an exception
            # because UI elements have been destroyed. Simply ignore it.
            pass
        evt.Skip()

    def OnNewsAndUpdates(self, evt):
        MyUpdateDialog.check_and_update(self.frame, self.pref, True)

    def OnDonate(self, evt):
        wx.LaunchDefaultBrowser(_("http://www.skeed.it/songpress#donate"))

    def OnFormatFont(self, evt):
        f = FontFaceDialog(
            self.frame,
            wx.ID_ANY,
            _("Songpress"),
            self.pref.format,
            self.pref.decorator,
            self.pref.decoratorFormat
        )
        if f.ShowModal() == wx.ID_OK:
            self.pref.SetFont(f.GetValue())
            self.SetFont()

    def OnTextFont(self, evt):
        data = wx.FontData()
        data.SetInitialFont(self.pref.format.wxFont)
        data.SetColour(self.pref.format.color)

        dialog = wx.FontDialog(self.frame, data)
        if dialog.ShowModal() == wx.ID_OK:
            retData = dialog.GetFontData()
            font = retData.GetChosenFont()
            face = font.GetFaceName()
            size = font.GetPointSize()
            color = retData.GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
            s = f"{{textfont:{face}}}{{textsize:{size}}}{{textcolour:{color}}}|"
            s += "{textfont}{textsize}{textcolour}"
            self.InsertWithCaret(s)

    def OnChordFont(self, evt):
        data = wx.FontData()
        data.SetInitialFont(self.pref.format.wxFont)
        data.SetColour(self.pref.format.color)

        dialog = wx.FontDialog(self.frame, data)
        if dialog.ShowModal() == wx.ID_OK:
            retData = dialog.GetFontData()
            font = retData.GetChosenFont()
            face = font.GetFaceName()
            size = font.GetPointSize()
            color = retData.GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
            s = f"{{chordfont:{face}}}{{chordsize:{size}}}{{chordcolour:{color}}}|"
            s += "{chordfont}{chordsize}{chordcolour}"
            self.InsertWithCaret(s)

    def OnTranspose(self, evt):
        t = MyTransposeDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
        if t.ShowModal() == wx.ID_OK:
            self.text.ReplaceTextOrSelection(t.GetTransposed())

    def OnSimplifyChords(self, evt):
        self.text.AutoChangeMode(True)
        t = self.text.GetTextOrSelection()
        notation = autodetectNotation(t, self.pref.notations)
        count, c, dc, e, de = findEasiestKey(t, self.pref.GetEasyChords(), notation)
        title = _("Simplify chords")
        if count > 0 and dc != de:
            msg = _("The key of your song, %s, is not the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
            msg += _("Do you want to transpose the key %s, which is the easiest one (difficulty: %.1f/5.0)?") % (e, 5 * de)
            d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
            if d.ShowModal() == wx.ID_YES:
                t = transposeChordPro(translateChord(c, notation), translateChord(e, notation), t, notation)
                self.text.ReplaceTextOrSelection(t)
        else:
            if count > 0:
                msg = _("The key of your song, %s, is already the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
            else:
                msg = _("Your song or current selection does not contain any chords.")
            d = wx.MessageDialog(self.frame, msg, title, wx.OK | wx.ICON_INFORMATION)
            d.ShowModal()
        self.text.AutoChangeMode(False)

    def OnChangeChordNotation(self, evt):
        t = MyNotationDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
        if t.ShowModal() == wx.ID_OK:
            self.text.ReplaceTextOrSelection(t.ChangeChordNotation())

    def OnNormalizeChords(self, evt):
        t = MyNormalizeDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
        if t.ShowModal() == wx.ID_OK:
            self.text.ReplaceTextOrSelection(t.NormalizeChords())

    def OnConvertTabToChordpro(self, evt):
        t = self.text.GetTextOrSelection()
        n = testTabFormat(t, self.pref.notations)
        if n is not None:
            self.text.ReplaceTextOrSelection(tab2ChordPro(t, n))

    def OnRemoveSpuriousBlankLines(self, evt):
        self.text.ReplaceTextOrSelection(removeSpuriousLines(self.text.GetTextOrSelection()))

    def OnOptions(self, evt):
        f = MyPreferencesDialog(self.frame, self.pref, easyChords)
        if f.ShowModal() == wx.ID_OK:
            self.text.SetFont(self.pref.editorFace, int(self.pref.editorSize))
            self.SetDefaultExtension(self.pref.defaultExtension)

    def StripSelection(self):
        """
        Update selection, moving blank characters out of it
        """
        s, e = self.text.GetSelection()
        mod = False
        while e > s and self.text.GetTextRange((ep := self.text.PositionBefore(e)), e).strip() == '':
            e = ep
            mod = True
        while s < e and self.text.GetTextRange(s, (sa := self.text.PositionAfter(s))).strip() == '':
            s = sa
            mod = True
        if mod:
            self.text.SetSelection(s, e)

    def InsertWithCaret(self, st):
        self.StripSelection()
        s, e = self.text.GetSelection()
        c = st.find('|')
        if c != -1:
            sel_text = self.text.GetSelectedText()
            self.text.ReplaceSelection(st[:c] + sel_text + st[c + 1:])
            self.text.SetSelection(s + c, e + c)
        else:
            self.text.ReplaceSelection(st)
            self.text.SetSelection(s + len(st), s + len(st))

    def OnTitle(self, evt):
        self.InsertWithCaret("{title:|}")

    def OnSubtitle(self, evt):
        self.InsertWithCaret("{subtitle:|}")

    def OnChord(self, evt):
        self.InsertWithCaret("[|]")

    def OnVerse(self, evt):
        label = wx.GetTextFromUser(
            _("Insert a label for verse, or press Cancel if you want to omit label."),
            _("Verse label"),
            "",
            self.frame,
        )
        self.InsertWithCaret("{Verse:%s}|" % label)

    def OnChorus(self, evt):
        default = self.pref.decoratorFormat.GetChorusLabel()
        label = wx.GetTextFromUser(
            _("Insert a label for chorus, or press Cancel if you want to omit label."),
            _("Chorus label"),
            default,
            self.frame,
        )
        if label == default:
            self.InsertWithCaret("{soc}\n|\n{eoc}\n")
        else:
            self.InsertWithCaret("{soc:%s}\n|\n{eoc}\n" % label)

    def OnComment(self, evt):
        self.InsertWithCaret("{c:|}")

    def OnLabelVerses(self, evt):
        self.pref.labelVerses = not self.pref.labelVerses
        self.CheckLabelVerses()

    def OnChorusLabel(self, evt):
        c = self.pref.decoratorFormat.GetChorusLabel()
        msg = _("Please enter a label for chorus")
        d = wx.TextEntryDialog(self.frame, msg, _("Songpress"), c)
        if d.ShowModal() == wx.ID_OK:
            c = d.GetValue()
            self.pref.SetChorusLabel(c)
            self.previewCanvas.Refresh(self.text.GetText())

    def OnNoChords(self, evt):
        self.pref.format.showChords = 0
        self.SetFont(True)

    def OnOneVerseForEachChordPattern(self, evt):
        self.pref.format.showChords = 1
        self.SetFont(True)

    def OnWholeSong(self, evt):
        self.pref.format.showChords = 2
        self.SetFont(True)

    def OnTextChanged(self, evt):
        self.AutoAdjust(evt.lastPos, evt.currentPos)

    def AutoAdjust(self, lastPos, currentPos):
        self.text.AutoChangeMode(True)
        t = self.text.GetTextRange(lastPos, currentPos)
        if self.pref.autoAdjustSpuriousLines:
            if testSpuriousLines(t):
                msg = _("It looks like there are spurious blank lines in the song.\n")
                msg += _("Do you want to try to remove them automatically?")
                title = _("Remove spurious blank lines")
                d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
                if d.ShowModal() == wx.ID_YES:
                    self.text.SetSelection(lastPos, currentPos)
                    t = removeSpuriousLines(t)
                    self.text.ReplaceSelection(t)
                    currentPos = self.text.GetCurrentPos()
        if self.pref.autoAdjustTab2Chordpro:
            n = testTabFormat(t, self.pref.notations)
            if n is not None:
                msg = _("It looks like your song is in tab format (i.e., chords are above the text).\n")
                msg += _("Do you want to convert it to ChordPro automatically?")
                title = _("Convert to ChordPro")
                d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
                if d.ShowModal() == wx.ID_YES:
                    self.text.SetSelection(lastPos, currentPos)
                    t = tab2ChordPro(t, n)
                    self.text.ReplaceSelection(t)
        if self.pref.autoAdjustEasyKey:
            notation = autodetectNotation(t, self.pref.notations)
            count, c, dc, e, de = findEasiestKey(t, self.pref.GetEasyChords(), notation)
            if count > 10 and dc != de:
                msg = _("The key of your song, %s, is not the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
                msg += _("Do you want to transpose the key %s, which is the easiest one (difficulty: %.1f/5.0)?") % (e, 5 * de)
                title = _("Simplify chords")
                d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
                if d.ShowModal() == wx.ID_YES:
                    self.text.SetSelection(lastPos, currentPos)
                    t = transposeChordPro(translateChord(c, notation), translateChord(e, notation), t, notation)
                    self.text.ReplaceSelection(t)
        self.text.AutoChangeMode(False)

    def SetFont(self, updateFontChooser=True):
        try:
            if updateFontChooser:
                self.fontChooser.SetValue(self.pref.format.face)
                self.showChordsChooser.SetValue(self.pref.format.showChords)
                ids = [self.noChordsMenuId, self.oneVerseForEachChordPatternMenuId, self.wholeSongMenuId]
                self.menuBar.Check(ids[self.pref.format.showChords], True)

            self.previewCanvas.Refresh(self.text.GetText())
        except wx._core.PyDeadObjectError:
            # When frame is closed, this method may still be executed, generating an exception
            # because UI elements have been destroyed. Simply ignore it.
            pass

    def CheckLabelVerses(self):
        self.formatToolBar.ToggleTool(self.labelVersesToolId, self.pref.labelVerses)
        self.formatToolBar.Refresh()
        self.menuBar.Check(self.labelVersesMenuId, self.pref.labelVerses)
        if self.pref.labelVerses:
            self.previewCanvas.SetDecorator(self.pref.decorator)
        else:
            self.previewCanvas.SetDecorator(SongDecorator())
        self.previewCanvas.Refresh(self.text.GetText())





import xmlrpc.client

import requests


class RequestsTransport(xmlrpc.client.Transport):
    def request(self, host, handler, data, verbose=False):
        # set the headers, including the user-agent
        headers = {
            "User-Agent": "my-user-agent",
            "Content-Type": "text/xml",
            "Accept-Encoding": "gzip"
        }
        url = "https://%s%s" % (host, handler)
        response = None
        try:
            response = requests.post(url, data=data, headers=headers)
            response.raise_for_status()
            return self.parse_response(response)
        except requests.RequestException as e:
            if response is None:
                raise xmlrpc.client.ProtocolError(url, 500, str(e), "")
            else:
                raise xmlrpc.client.ProtocolError(url, response.status_code, str(e), response.headers)

    def parse_response(self, resp):
        """
        Parse the xmlrpc response.
        """
        p, u = self.getparser()
        p.feed(resp.text)
        p.close()
        return u.close()


###############################################################
# Name:             Editor.py
# Purpose:     Subclass of StyledTextControl providing songpress
#            editor: loading, saving and syntax hilighting
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2009-01-16
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import codecs

from wx.stc import *
import wx.lib, wx.lib.newevent

from .SDIMainFrame import *
from .SongTokenizer import *
from . import Transpose


EventTextChanged, EVT_TEXT_CHANGED = wx.lib.newevent.NewEvent()


def get_text_from_clipboard():
    """
    Return text in clipboard, or None
    """
    if wx.TheClipboard.IsOpened():
        return None
    wx.TheClipboard.Open()
    do = wx.TextDataObject()
    if not wx.TheClipboard.GetData(do):
        return None
    wx.TheClipboard.Close()
    return do.GetText()


class Editor(StyledTextCtrl):
    def __init__(self, spframe, interactive=True, frame=None):
        self.frame = spframe.frame if frame is None else frame
        StyledTextCtrl.__init__(self, self.frame)
        self.spframe = spframe
        self.interactive = interactive
        if self.interactive:
            self.lastPos = 0
            self.frame.Bind(EVT_STC_CHANGE, self.OnTextChange, self)
            self.frame.Bind(EVT_STC_UPDATEUI, self.OnUpdateUI, self)
            self.Bind(EVT_STC_DOUBLECLICK, self.OnDoubleClick, self)
            self.Bind(wx.EVT_CHAR, self.OnChar, self)
            # self.Bind(EVT_STC_CLIPBOARD_COPY, self.frame.OnCopy, self)
            self.autoChangeMode = False
        self.frame.Bind(EVT_STC_STYLENEEDED, self.OnStyleNeeded, self)
        self.STC_STYLE_NORMAL = 11
        self.STC_STYLE_CHORD = 12
        self.STC_STYLE_COMMAND = 13
        self.STC_STYLE_ATTR = 14
        self.STC_STYLE_CHORUS = 15
        self.STC_STYLE_COMMENT = 16
        self.SetFont("Lucida Console", 12)
        self.SetLexer(STC_LEX_CONTAINER)
        self.StyleSetForeground(self.STC_STYLE_NORMAL, wx.Colour(0, 0, 0))
        self.StyleSetForeground(self.STC_STYLE_CHORUS, wx.Colour(0, 0, 0))
        self.StyleSetForeground(self.STC_STYLE_CHORD, wx.Colour(255, 0, 0))
        self.StyleSetForeground(self.STC_STYLE_COMMAND, wx.Colour(0, 0, 255))
        self.StyleSetForeground(self.STC_STYLE_ATTR, wx.Colour(0, 128, 0))
        self.StyleSetForeground(self.STC_STYLE_COMMENT, wx.Colour(128, 128, 128))
        self.StyleSetBold(self.STC_STYLE_CHORUS, True)
        #Dummy "token": we artificially replace every normalToken into a chorusToken when we are
        #inside chorus.  Then, we can associate the chorus style in self.tokenStyle dictionary.
        self.chorusToken = 'chorusToken'
        self.tokenStyle = {
            SongTokenizer.openCurlyToken: self.STC_STYLE_COMMAND,
            SongTokenizer.closeCurlyToken: self.STC_STYLE_COMMAND,
            SongTokenizer.normalToken: self.STC_STYLE_NORMAL,
            SongTokenizer.commandToken: self.STC_STYLE_COMMAND,
            SongTokenizer.attrToken: self.STC_STYLE_ATTR,
            SongTokenizer.chordToken: self.STC_STYLE_CHORD,
            SongTokenizer.closeChordToken: self.STC_STYLE_CHORD,
            SongTokenizer.colonToken: self.STC_STYLE_COMMAND,
            SongTokenizer.commentToken: self.STC_STYLE_COMMENT,
            self.chorusToken: self.STC_STYLE_CHORUS
        }
        #self.chorus[i] == True iff, at the end of line i, we are still in chorus (i.e. bold) mode
        self.chorus = []

    def SetFont(self, face, size):
        font = wx.Font(
            size,
            wx.FONTFAMILY_DEFAULT,
            wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_NORMAL,
            faceName = face
        )
        self.StyleSetFont(STC_STYLE_DEFAULT, font)
        #I don't know why, but the following line is necessary in order to make
        #the font bold
        self.StyleSetFont(self.STC_STYLE_CHORUS, font)

    def New(self):
        self.ClearAll()

    def Open(self):
        self.ClearAll()
        self.LoadFile(self.spframe.document)

    def Save(self):
        #self.SaveFile(self.spframe.document)
        t = self.GetText()
        f = codecs.open(self.spframe.document, 'w', 'utf-8')
        f.write(t)
        f.close()

    def GetTextOrSelection(self):
        start, end = self.GetSelection()
        if start == end:
            return self.GetText()
        return self.GetSelectedText()

    def ReplaceTextOrSelection(self, text):
        start, end = self.GetSelection()
        if start == end:
            self.SetText(text)
        else:
            self.ReplaceSelection(text)

    def GetChordUnderCursor(self):
        """Return info about chord under cursor, or None

        Return a 3-tuple: (begin, end, chord)
            begin: position before open bracket
            end: position after close bracket
            chord: chord, without brackets
        """
        pos, dummy = self.GetSelection()
        char = ""
        if self.GetSelectedText().find('[') != -1:
            start = pos
            while char != '[':
                next = self.PositionAfter(start)
                char = self.GetTextRange(start, next)
                start = next
            start = self.PositionBefore(start)
        else:
            start = self.PositionBefore(pos)
            while start >= 0 and char != '[' and char != '\n' and char != ']':
                char = self.GetTextRange(start, self.PositionAfter(start))
                if start == 0:
                    start = -1
                else:
                    start = self.PositionBefore(start)
            if start >= 0:
                start = self.PositionAfter(start)
            else:
                start = 0
        if char == '[':
            end = self.PositionBefore(pos)
            l = self.GetLength()
            while end < l and char != ']' and char != '\n':
                char = self.GetTextRange(end, self.PositionAfter(end))
                end = self.PositionAfter(end)
            if char == ']':
                return (start, end, self.GetTextRange(self.PositionAfter(start), self.PositionBefore(end)))
        return None

    def SelectNextChord(self):
        dummy, pos = self.GetSelection()
        n = self.GetLength()
        c = ''
        while pos < n:
            while pos < n and c != '[':
                c = self.GetTextRange(pos, self.PositionAfter(pos))
                pos = self.PositionAfter(pos)
            if c == '[':
                start = pos
                while pos < n and c != ']' and c != '\n':
                    c = self.GetTextRange(pos, self.PositionAfter(pos))
                    pos = self.PositionAfter(pos)
                if c == ']':
                    self.SetSelection(self.PositionBefore(start), pos)
                    return True
        return False

    def SelectPreviousChord(self):
        pos, dummy = self.GetSelection()
        c = ''
        done = False
        while not done:
            while not done and c != ']':
                c = self.GetTextRange(pos, self.PositionAfter(pos))
                if pos <= 0:
                    done = True
                else:
                    pos = self.PositionBefore(pos)
            if c == ']':
                end = pos
                while not done and c != '[' and c != '\n':
                    c = self.GetTextRange(pos, self.PositionAfter(pos))
                    if pos <= 0:
                        pos = -1
                        done = True
                    else:
                        pos = self.PositionBefore(pos)
                if pos == -1:
                    pos = 0
                else:
                    pos = self.PositionAfter(pos)
                if c == '[':
                    self.SetSelection(pos, self.PositionAfter(self.PositionAfter(end)))
                    return True
        return False

    def RemoveChordsInSelection(self):
        self.ReplaceSelection(Transpose.removeChords(self.GetSelectedText()))
        
    def CopyOnlyText(self):
        text = Transpose.removeChordPro(self.GetSelectedText())
        c = wx.TheClipboard
        if c.Open():
            c.SetData(wx.TextDataObject(text))
            c.Close()

    def OnTextChange(self, evt):
        if self.interactive:
            self.spframe.SetModified()
            self.spframe.TextUpdated()
            if not self.autoChangeMode:
                currentPos = self.GetCurrentPos()
                if currentPos - self.lastPos > 2:
                    evt = EventTextChanged(lastPos=self.lastPos, currentPos=currentPos)
                    wx.PostEvent(self.frame, evt)

    def OnUpdateUI(self, evt):
        self.lastPos = self.GetCurrentPos()
        evt.Skip(False)

    def OnDoubleClick(self, evt):
        t = self.GetChordUnderCursor()
        if t is not None:
            self.SetSelection(t[0], t[1])
        evt.Skip()

    def OnChar(self, evt):
        t = self.GetSelectedText()
        if t != '' and t[0] == '[' and t[-1] == ']':
            c = evt.GetKeyCode()
            if (c >= 65 and c <= 90) or (c >= 97 and c <= 122):
                s, e = self.GetSelection()
                self.SetSelection(self.PositionAfter(s), self.PositionBefore(e))
        evt.Skip()

    def AutoChangeMode(self, acm):
        self.autoChangeMode = acm
        
    def PasteChords(self):
        src = get_text_from_clipboard()
        if src is None:
            return
        self.BeginUndoAction()
        start, end = self.GetSelection()
        if start == end:
            l = self.LineFromPosition(start)
            end = self.PositionFromLine(l + len(src.splitlines()))
            if end == -1:
                end = self.GetLength()
            else:
                end = self.PositionBefore(end)
        prev = self.PositionBefore(end)
        while end > start and self.GetCharAt(prev) in [10, 13]:
            end = prev
            prev = self.PositionBefore(end)
        self.SetSelection(start, end)
        sel = self.GetSelectedText()
        if sel == "":
            sel = " "
        self.ReplaceSelection(Transpose.pasteChords(src, sel))
        self.EndUndoAction()

    def OnStyleNeeded(self, evt):
        end = evt.GetPosition()
        pos = self.GetEndStyled()
        ln = self.LineFromPosition(pos)
        start = self.PositionFromLine(ln)
        if ln > 0 and len(self.chorus) > ln-1:
            bold = self.chorus[ln-1]
        else:
            bold = False
        #changedBold is True iff self.chorus has changed for the current line, and thus we need
        #to process (at least) another line
        changedBold = False
        lc = self.GetLineCount()
        while (changedBold and ln < lc) or start < end:
            self.StartStyling(start)
            l = self.GetLine(ln)
            tkz = SongTokenizer(l)
            for tok in tkz:
                n = len(tok.content.encode('utf-8'))
                t = tok.token
                if bold and t == SongTokenizer.normalToken:
                    t = self.chorusToken
                self.SetStyling(n, self.tokenStyle[t])
                if t == SongTokenizer.commandToken:
                    if tok.content.upper() == 'SOC':
                        bold = True
                    elif tok.content.upper() == 'EOC':
                        bold = False
            if len(self.chorus) > ln:
                changedBold = self.chorus[ln] != bold
                self.chorus[ln] = bold
            else:
                self.chorus.append(bold)
                changedBold = False
            ln = ln + 1
            start = self.PositionFromLine(ln)





# -*- coding: utf-8 -*- 

###########################################################################
## Python code generated with wxFormBuilder (version Aug 25 2009)
## http://www.wxformbuilder.org/
##
## PLEASE DO "NOT" EDIT THIS FILE!
###########################################################################

import wx

_ = wx.GetTranslation

# -*- coding: utf-8 -*- 


###########################################################################
## Class NormalizeDialog
###########################################################################

class NormalizeDialog ( wx.Dialog ):
    
    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = _(u"Cleanup Chords"), pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER )
        
        self.SetSizeHintsSz( wx.DefaultSize, wx.DefaultSize )
        
        bSizer1 = wx.BoxSizer( wx.VERTICAL )
        
        bSizer2 = wx.BoxSizer( wx.HORIZONTAL )
        
        self.label1 = wx.StaticText( self, wx.ID_ANY, _(u"Notation:"), wx.DefaultPosition, wx.DefaultSize, 0 )
        self.label1.Wrap( -1 )
        bSizer2.Add( self.label1, 0, wx.ALL, 5 )
        
        fromNotationChoices = []
        self.fromNotation = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, fromNotationChoices, 0 )
        self.fromNotation.SetSelection( 0 )
        bSizer2.Add( self.fromNotation, 1, wx.ALL, 5 )
        
        
        bSizer1.Add( bSizer2, 1, wx.EXPAND, 5 )
        
        m_sdbSizer1 = wx.StdDialogButtonSizer()
        self.m_sdbSizer1OK = wx.Button( self, wx.ID_OK )
        m_sdbSizer1.AddButton( self.m_sdbSizer1OK )
        self.m_sdbSizer1Cancel = wx.Button( self, wx.ID_CANCEL )
        m_sdbSizer1.AddButton( self.m_sdbSizer1Cancel )
        m_sdbSizer1.Realize();
        
        bSizer1.Add( m_sdbSizer1, 0, wx.ALL|wx.EXPAND, 5 )
        
        
        self.SetSizer( bSizer1 )
        self.Layout()
        bSizer1.Fit( self )
        
        # Connect Events
        self.fromNotation.Bind( wx.EVT_CHOICE, self.OnFromNotation )
    
    def __del__( self ):
        pass
    
    
    # Virtual event handlers, overide them in your derived class
    def OnFromNotation( self, event ):
        event.Skip()
    




###############################################################
# Name:             MyUpdateDialog.py
# Purpose:     Dialog to select updates to apply
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2010-12-14
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################
from xmlrpc.client import Server
import datetime
import platform
import traceback
import logging

import wx.lib.delayedresult as delayedresult
import wx

from .UpdateDialog import *
from .Globals import glb
from .MyUpdatePanel import MyUpdatePanel
from .proxiedxmlrpclib import RequestsTransport


from . import i18n


_ = wx.GetTranslation

def check_and_update(parent, preferences, force=False):
    """
    Check for updates, and interact with user.
    
    Update mode is automatic if force=False, else manual.
    
    parent: parent window of any created modal dialogs
    """
    def check_for_updates():
        """
        Check and return available updates.
        
        Honour the frequency of update checking unless force is True.
        
        Return list of pair (version, description),
        where description is an html-formatted text.
        Raise an exception if anything goes wrong.
        """
        if(
            force
            or preferences.updateLastCheck is None
            or datetime.datetime.now() > preferences.updateLastCheck + datetime.timedelta(days=preferences.updateFrequency)
        ):
            tr = RequestsTransport()
            s = Server(preferences.updateUrl, transport=tr)
            # method returns a dictionary with those keys:
            # 'new_url': if defined, new url of the webservice
            # 'updates': list of 3-tuples (version, description, downloadUrl)
            version = glb.VERSION
            version_lst = version.split(".")
            if len(version_lst) > 3 or not version_lst[-1].isdigit():
                # Prerelease
                version = ".".join(version_lst[:2]) + ".-1"
            print(version)
            u = s.checkForUpdates(
                version,
                preferences.updateFrequency,
                platform.system(),
                platform.architecture(),
                platform.platform(),
                platform.python_version(),
                wx.version(),
                i18n.getLang(),
            )
            u2 = [x for x in u['updates'] if x[0] not in preferences.ignoredUpdates]
            preferences.updateLastCheck = datetime.datetime.now()
            if 'new_url' in u:
                preferences.updateUrl = u['new_url']
            return u2
        return []

    def consume_updates(dr):
        try:
            u = dr.get()
            if len(u) > 0:
                d = MyUpdateDialog(parent, preferences, u)
                d.ShowModal()
            elif force:
                d = wx.MessageDialog(
                    parent,
                    _("Your version of Songpress is up to date."),
                    _("Songpress"),
                    wx.OK | wx.ICON_INFORMATION
                )
                d.ShowModal()
        except Exception as e:
            logging.error(traceback.format_exc())
            if force:
                d = wx.MessageDialog(
                    parent,
                    _("There was an error while checking for updates.\nPlease try again later."),
                    _("Update error"),
                    wx.OK | wx.ICON_ERROR
                )
                d.ShowModal()

    delayedresult.startWorker(consume_updates, check_for_updates)


class MyUpdateDialog(UpdateDialog):
    def __init__(self, parent, preferences, updates):
        UpdateDialog.__init__(self, parent)
        self.preferences = preferences
        self.updates = updates
        self.updatePanels = set()
        for u in updates:
            up = MyUpdatePanel(self, preferences, u[0], u[1], u[2])
            self.updatePanels.add(up)
            self.updatesSizer.Add(up, 1, wx.EXPAND, 5 )
        self.SetSize((400, 400))
        
    def RemoveChild(self, child):
        self.updatesSizer.Layout()
        self.updatePanels.remove(child)
        if len(self.updatePanels) == 0:
            self.Show(False)
            self.Destroy()
            
    def OnDonate(self, evt):
        wx.LaunchDefaultBrowser(_("https://www.skeed.it/songpress#donate"))
        evt.Skip()




# -*- coding: iso-8859-1 -*-

###############################################################
# Name:             EditDistance.py
# Purpose:     Minimum Edit Distance algorithm
# Code from: http://www.cs.colorado.edu/~martin/csci5832/edit-dist-blurb.html
# Created:     2013-08-14
##############################################################


def substCost(x,y):
    if x == y: return 0
    else: return 1
    
def insertCost(x):
    return 1

def deleteCost(x):
    return 1

def  minEditDist(target, source):
    n = len(target)
    m = len(source)

    distance = [[0 for i in range(m+1)] for j in range(n+1)]

    for i in range(1,n+1):
        distance[i][0] = distance[i-1][0] + insertCost(target[i-1])

    for j in range(1,m+1):
        distance[0][j] = distance[0][j-1] + deleteCost(source[j-1])

    for i in range(1,n+1):
        for j in range(1,m+1):
           distance[i][j] = min(
                    distance[i-1][j]+1,
                    distance[i][j-1]+1,
                    distance[i-1][j-1]+substCost(source[j-1],target[i-1])
                )
    return distance[n][m]



###############################################################
# Name:             SongFormat.py
# Purpose:     Song format options
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2009-02-24
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import wx


class AttributeMonitor(object):
    def __init__(self, handlers):
        object.__init__(self)
        self.am_started = False
        self.am_handlers = handlers

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if name != 'am_started' and self.am_started:
            for h in self.am_handlers:
                if name in h.names:
                    #print name
                    h.OnSetAttr(self, name, value)

    def Am_Start(self):
        self.am_started = True


class FontChangeHandler(object):
    names = set(("face", "size", "bold", "italic", "underline"))
    @staticmethod
    def OnSetAttr(format, name, value):
        format.UpdateWxFont()


class FontFormat(AttributeMonitor):
    def __init__(self, ff=None):
        AttributeMonitor.__init__(self, (FontChangeHandler, ))
        # Definable by user
        if ff == None:
            self.face = "Arial"
        else:
            self.face = ff.face
        self.size = 12 if ff is None else ff.size
        self.bold = False if ff is None else ff.bold
        self.italic = False if ff is None else ff.italic
        self.underline = False  if ff is None else ff.underline
        self.color = '#000000' if ff is None else ff.color
        # Auto set
        self.wxFont = None
        self.UpdateWxFont()
        # Start monitor
        self.Am_Start()

    def UpdateWxFont(self):
        #print("Font Updated, face = " + self.face)
        if self.italic:
            style = wx.FONTSTYLE_ITALIC
        else:
            style = wx.FONTSTYLE_NORMAL
        if self.bold:
            weight = wx.FONTWEIGHT_BOLD
        else:
            weight = wx.FONTWEIGHT_NORMAL
        self.wxFont = wx.Font(int(self.size), wx.FONTFAMILY_DEFAULT, style, weight, self.underline, self.face)


class ParagraphFormat(FontFormat):
    def __init__(self, ff=None):
        FontFormat.__init__(self, ff)
        self.leftMargin = 0 if ff is None else ff.leftMargin
        self.topMargin = 12 if ff is None else ff.topMargin
        self.bottomMargin = 0 if ff is None else ff.bottomMargin
        self.chordSpacing = 0.8 if ff is None else ff.chordSpacing
        self.textSpacing = 1 if ff is None else ff.textSpacing
        self.chord = FontFormat(ff.chord) if ff is not None else FontFormat()
        self.comment = FontFormat(ff.comment) if ff is not None else FontFormat()
        if ff is None:
            self.chord.size = self.size * 0.9
            self.chord.italic = True
            self.chord.bold = False
            self.comment.italic = True


class SongFormat(ParagraphFormat):
    def __init__(self, ff=None):
        ParagraphFormat.__init__(self, ff)
        self.verse = []
        self.chorus = ParagraphFormat(ff.chorus) if ff is not None else ParagraphFormat()
        self.title = ParagraphFormat(ff.title) if ff is not None else ParagraphFormat()
        self.subtitle = ParagraphFormat(ff.subtitle) if ff is not None else ParagraphFormat()
        self.blockSpacing = 1 if ff is None else ff.blockSpacing
        # showChords:
        # 0. None
        # 1. First verse and chorus
        # 2. Entire song
        self.showChords = 1 if ff is None else ff.showChords
        if ff is None:
            self.chorus.bold = True
            self.chorus.underline = False
            self.title.bold = True
            self.title.underline = True
            self.subtitle.size = self.title.size * 0.95
            self.subtitle.italic = True

    def AddVerse(self):
        self.verse.append(ParagraphFormat(self))

    def InitVerses(self):
        self.verse = []


###############################################################
# Name:             MyDecoSlider.py
# Purpose:     Slider with a decorated range indicator
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2010-06-02
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import wx

from .DecoSlider import *

class MyDecoSlider(DecoSlider):
    def __init__(self, parent):
        DecoSlider.__init__(self, parent)

    def OnPaint(self, event):
        dc = wx.PaintDC(self.panel)
        w, h = dc.GetSize()
        dc.Clear()
        green =    wx.Colour(0, 255, 0)
        brush = wx.Brush(green, wx.SOLID)
        dc.SetBrush(brush)
        dc.DrawPolygon([wx.Point(0, 0), wx.Point(w - 1, 0), wx.Point(w - 1, h - 1)])

    def OnSize(self, event):
        self.panel.Refresh()



###############################################################################
# Name: dev_tool.py                                                           #
# Purpose: Provides logging and error tracking utilities                      #
# Original author: Cody Precord <cprecord@editra.org>                         #
# Modified by Luca Allulli (just minor adaptations)                           #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

""" Editra Development Tools
Tools and Utilities for debugging and helping with development of Editra.
@summary: Utility function for debugging the editor

"""
__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: dev_tool.py 61794 2009-08-31 04:03:04Z CJP $"
__revision__ = "$Revision: 61794 $"

#-----------------------------------------------------------------------------#
# Imports
import os
import sys
import traceback
import urllib.request as urllib2
import webbrowser
import wx

# Editra Libraries
#import ed_glob
#import ed_msg
#import eclib
#from ebmlib import IsUnicode
from . import errdlg
from .Globals import glb

#-----------------------------------------------------------------------------#
# Globals

class EdErrorDialog(errdlg.ErrorDialog):
    def __init__(self, msg):
        errdlg.ErrorDialog.__init__(self, None, title="Error Report", message=msg)

        # Setup
        self.SetDescriptionLabel("Error: Something unexpected happend\n"
                                   "Help improve %s by clicking on "
                                   "Report Error\nto send the Error "
                                   "Traceback shown below." % (glb.PROG_NAME,))

    def Abort(self):
        """Abort the application"""
        sys.exit(1)

    def GetProgramName(self):
        """Get the program name to display in error report"""
        return "%s Version: %s" % (glb.PROG_NAME, glb.VERSION)

    def Send(self):
        """Send the error report"""
        msg = "mailto:%s?subject=Error Report&body=%s"
        addr = glb.BUG_REPORT_ADDRESS
        if wx.Platform != '__WXMAC__':
            body = urllib2.quote(self.err_msg)
        else:
            body = self.err_msg
        msg = msg % (addr, body)
        msg = msg.replace("'", '')
        webbrowser.open(msg)

#-----------------------------------------------------------------------------#

def ExceptionHook(exctype, value, trace):
    """Handler for all unhandled exceptions
    @param exctype: Exception Type
    @param value: Error Value
    @param trace: Trace back info

    """
    # Format the traceback
    exc = traceback.format_exception(exctype, value, trace)
    exc.insert(0, u"*** %s ***%s" % (errdlg.TimeStamp(), os.linesep))
    ftrace = u"".join(exc)

    # Ensure that error gets raised to console as well
    print(ftrace)

    # If abort has been set and we get here again do a more forcefull shutdown
    if EdErrorDialog.ABORT:
        os._exit(1)

    # Prevent multiple reporter dialogs from opening at once
    if not EdErrorDialog.REPORTER_ACTIVE and not EdErrorDialog.ABORT:
        dlg = EdErrorDialog(ftrace)
        dlg.ShowModal()
        dlg.Destroy()

#-----------------------------------------------------------------------------#



"""Framework providing common functionality for SDI Main Frames

The module provides an (abstract) class loads an XRC resource describing a SDI Frame.
It expects to find some menu elements, characterized by their XRC name:
    - new
    - open
    - save
    - saveAs
    - exit
    - about
"""

###############################################################
# Name:             SDIMainFrame.py
# Purpose:     Abstract class for SDI Main Frames
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2009-01-16
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import os
import sys
import platform

import wx
# import wx.aui as aui
from wx.lib.agw import aui
from wx import xrc

from . import i18n


_ = wx.GetTranslation

class SDIDropTarget(wx.FileDropTarget):
    def __init__(self, sdi):
        wx.FileDropTarget.__init__(self)
        self.sdi = sdi

    def OnDropFiles(self, x, y, arr):
        return self.sdi.OnDropFiles(arr)


class SDIMainFrame(wx.FileDropTarget):
    """Abstract class. Override methods New, Open, Save"""
    ###UI generation###

    def __init__(
        self,
        res,
        frameName='MainFrame',
        appName='SDIApp',
        authorName='Nobody',
        docType='document',
        docExt='txt',
        appLongName=None,
        icon=None,
        version="1.0",
        url="",
        copyright = "",
        licensing = "",
        thanks = "",
        importFormats = [] # List of tuples: (format name, [extensions])
    ):
        self.config = wx.Config.Get()
        self.res = res
        i18n.localizeXrc('xrc/songpress.xrc')
        self.appName = appName
        self.appLongName = self.appName if appLongName == None else appLongName
        self.authorName = authorName
        self.modified = False
        self.version = version
        self.url = url
        self.copyright = copyright
        self.licensing = licensing
        self.thanks = thanks
        self.document = ''
        self.docType = docType
        self.docExt = docExt
        self.importFormats = importFormats
        wx.Config.Set(self.config)
        self.frame = self.res.LoadFrame(None, frameName)
        if icon != None:
            self.frame.SetIcon(wx.Icon(icon, wx.BITMAP_TYPE_ICO))
        self.BindMenu()
        self.frame.Bind(wx.EVT_CLOSE, self.OnClose, self.frame)
        dt = SDIDropTarget(self)
        self.frame.SetDropTarget(dt)
        self.UpdateTitle()
        self._mgr = aui.AuiManager(self.frame, aui.AUI_MGR_ALLOW_FLOATING | aui.AUI_MGR_LIVE_RESIZE | aui.AUI_MGR_HINT_FADE)
        self._mgr.Bind(aui.EVT_AUI_PANE_CLOSE, self.OnPaneClose)
        self.menuBar = self.frame.GetMenuBar()
        self.panesByMenu = {}
        self.menusByPane = {}
        self.RetrieveRecentFileList()

    def Bind(self, event, handler, xrcname):
        """Bind an event, coming from a control by xrc name, to a handler"""
        self.frame.Bind(event, handler, id=xrc.XRCID(xrcname))

    def BindMenu(self):
        """Bind a menu item, by xrc name, to a handler"""
        def Bind(handler, xrcname):
            self.Bind(wx.EVT_MENU, handler, xrcname)

        Bind(self.OnNew, 'new')
        Bind(self.OnOpen, 'open')
        Bind(self.OnSave, 'save')
        Bind(self.OnSaveAs, 'saveAs')
        Bind(self.OnExit, 'exit')
        Bind(self.OnAbout, 'about')
        
    def SetDefaultExtension(self, ext):
        self.docExt = ext

    def OnNew(self, evt):
        """Menu handler for File->New"""
        if self.AskSaveModified():
            self.document = ''
            self.New()
            self.modified = False
            self.UpdateTitle()

    def OnOpen(self, evt):
        """Menu handler for File->Open"""
        if self.AskSaveModified():
            if self.importFormats == []:
                filter = _("%s files (*.%s)|%s") % (
                        self.docExt,
                        self.docExt,
                        self.docExt,
                    )
            else:
                filter = "|".join(["%s|%s" % (x[0], ";".join(["*." + y for y in x[1]])) for x in self.importFormats])
            filter += _("|All files (*.*)|*.*")
            dlg = wx.FileDialog(
                self.frame,
                _("Open file"),
                "",
                "",
                filter,
                wx.FD_OPEN
            )

            if dlg.ShowModal() == wx.ID_OK:
                fn = dlg.GetPath()
                if os.path.isfile(fn):
                    self.document = fn
                    self.Open()
                    self.UpdateRecentFileList(fn)
                    self.modified = False
                    self.UpdateTitle()
                else:
                    msg = _("File \"%s\" does not exist.") % (fn, )
                    d = wx.MessageDialog(
                        self.frame,
                        msg,
                        self.appLongName,
                        wx.OK | wx.ICON_ERROR
                    )
                    d.ShowModal()

    def OnSave(self, evt):
        """Menu handler for File->Save"""
        self.SaveFile()

    def OnSaveAs(self, evt):
        """Menu handler for File->Save As"""
        if self.AskSaveFilename():
            self.SaveFile()

    def OnExit(self, evt):
        """Menu handler for File->Exit"""
        self.frame.Close()

    def OnAbout(self, evt):
        """Menu handler for ?->About"""
        msg = _("%s version %s\n%s\n\n%s\n\n%s\n\n%s") % (
            self.appLongName,
            self.version,
            self.url,
            self.copyright,
            self.licensing,
            self.thanks
        )
        wx.MessageBox(msg, _('About ') + self.appLongName)

    def OnDropFiles(self, arr):
        """Handler for drop action: opens the dropped file, if it is exactly one"""
        if len(arr) == 1:
            fn = arr[0]
            if os.path.isfile(fn) and self.AskSaveModified():
                self.document = fn
                self.Open()
                self.UpdateRecentFileList(fn)
                self.modified = False
                self.UpdateTitle()
                return True
        return False

    def OnClose(self, evt):
        """Handler for windows close event"""
        if self.AskSaveModified(evt.CanVeto()):
            self.config.SetPath('/SDIMainFrame')
            self.config.Write("Version", self.version)
            p = self._mgr.SavePerspective()
            self.config.Write("Perspective", p)
            self.SavePreferences()
            self._mgr.UnInit()
            del self._mgr
            self.frame.Destroy()
        else:
            evt.Veto()

    ###Ordinary methods###

    def SetModified(self, m = True):
        """Set the modified flag, if main document is modified"""
        self.modified = m
        self.UpdateTitle()

    def UpdateTitle(self):
        """Updates form title; to be called when the filename or the modified status changes"""
        if self.modified:
            mod = '* '
        else:
            mod = ''
        if self.document == '':
            doc = _('Untitled')
        else:
            doc = os.path.basename(self.document)
            (doc, ext) = os.path.splitext(doc)
        self.frame.SetTitle(u"%s%s - %s" % (mod, doc, self.appLongName))

    def AskSaveModified(self, canCancel = True):
        """If file has been modified, propose to save changes. Return False if cancelled, True otherwise"""
        if not self.modified:
            return True

        if canCancel:
            cc = wx.CANCEL
        else:
            cc = 0

        d = wx.MessageDialog(self.frame, _("Your %s has been modified. Do you want to save it?") % (self.docType), self.appLongName, wx.YES_NO | wx.ICON_QUESTION | cc)
        res = d.ShowModal()
        if res == wx.ID_CANCEL:
            return False
        elif res == wx.ID_NO:
            return True
        else: #wxID_YES
            return self.SaveFile();

    def AskSaveFilename(self):
        """Ask and updates the filename (without saving); return False if user cancels, True otherwise"""
        leave = False;
        consensus = False;
        while not leave:
            dlg = wx.FileDialog(
                self.frame,
                _("Choose a name for the file"),
                "",
                "",
                _("%s files (*.%s)|*.%s|All files (*.*)|*.*") % (
                        self.docExt,
                        self.docExt,
                        self.docExt,
                    ),
                wx.FD_SAVE
            )

            if dlg.ShowModal() == wx.ID_OK:

                fn = dlg.GetPath()
                if os.path.isfile(fn):
                    msg = _("File \"%s\" already exists. Do you want to overwrite it?") % (fn, )
                    d = wx.MessageDialog(
                        self.frame,
                        msg,
                        self.appLongName,
                        wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION
                    )
                    res = d.ShowModal()
                    if res == wx.ID_CANCEL:
                        leave = True
                        consensus = False
                    elif res == wx.ID_NO:
                        leave = False
                        consensus = False
                    else: #wxID_YES
                        leave = True
                        consensus = True
                else:
                    leave = True
                    consensus = True

            else:
                leave = True
                consensus = False

        if consensus:
            if platform.system() == 'Linux':
                # Since wxPython file dialog in Linux does not add default extension
                # when it has not been specified by user, we add it ourselves
                pref, ext = os.path.splitext(fn)
                if ext == '':
                    fn = '%s.%s' % (fn, self.docExt)
            self.document = fn
            self.UpdateTitle()
            return wx.ID_OK
        else:
            return False

    def SaveFile(self):
        """Save file, asking for file name if necessary. Return False if user cancels, True otherwise"""
        if self.document == '':
            if not self.AskSaveFilename():
                return False
        self.Save()
        self.UpdateRecentFileList(self.document)
        self.modified = False
        self.UpdateTitle()
        return True

    def New(self):
        """To be overridden: create a blank document"""
        pass

    def Open(self):
        """To be overridden: open a document"""
        pass

    def Save(self):
        """To be overridden: save a document"""
        return True

    def SavePreferences(self):
        """To be overridden: save preferences to wxConfig object"""
        pass

    def AddMainPane(self, window):
        self._mgr.AddPane(window, aui.AuiPaneInfo().CenterPane().Name('_main'))

    def AddPane(self, window, info, caption, menuName):
        self._mgr.AddPane(window, info.Name(menuName).Caption(caption))
        menuid = xrc.XRCID(menuName)
        self.panesByMenu[menuid] = menuName
        self.menusByPane[menuName] = menuid
        self.Bind(wx.EVT_MENU, self.OnTogglePaneView, menuName)

    def OnTogglePaneView(self, evt):
        status = evt.GetInt()
        menu = evt.GetId()
        pane = self._mgr.GetPane(self.panesByMenu[menu])
        pane.Show(status)
        self._mgr.Update()

    def OnPaneClose(self, evt):
        pane = evt.GetPane()
        menuid = self.menusByPane[pane.name]
        self.menuBar.Check(menuid, False)

    def RetrieveRecentFileList(self):
        self.recentMenuBase = 800
        self.config.SetPath('/SDIMainFrame/RecentFiles')
        self.recentFiles = []
        for i in range(1, 5):
            f = self.config.Read(str(i))
            if f:
                self.recentFiles.append(f)
            self.frame.Bind(wx.EVT_MENU, self.OnRecentFile, id = self.recentMenuBase + i)
        self.UpdateRecentFileMenu()

    def EmptyRecentFileMenu(self):
        fileMenu = self.menuBar.GetMenu(0)
        for i in range(1, len(self.recentFiles)+1):
            fileMenu.Remove(self.recentMenuBase + i)

    def UpdateRecentFileMenu(self):
        i = 1
        fileMenu = self.menuBar.GetMenu(0)
        self.config.SetPath('/SDIMainFrame/RecentFiles')
        for k in self.recentFiles:
            d, f = os.path.split(k)
            if len(d) > 25:
                d = d[:25] + "..."
            fileMenu.Append(self.recentMenuBase + i, os.path.join(d, f))
            self.config.Write(str(i), k)
            i += 1

    def UpdateRecentFileList(self, name):
        self.EmptyRecentFileMenu()
        if name in self.recentFiles:
            self.recentFiles.remove(name)
        self.recentFiles.insert(0, name)
        if len(self.recentFiles) > 4:
            self.recentFiles.pop()
        self.UpdateRecentFileMenu()

    def OnRecentFile(self, evt):
        i = evt.GetId() - self.recentMenuBase
        fn = self.recentFiles[i-1]
        if os.path.isfile(fn) and self.AskSaveModified():
            self.document = fn
            self.Open()
            self.UpdateRecentFileList(fn)
            self.modified = False
            self.UpdateTitle()

    def FinalizePaneInitialization(self):
        self.config.SetPath('/SDIMainFrame')
        v = self.config.Read("Version", "0.0")
        vs = v.split('.')
        svs = self.version.split('.')
        p = False
        if vs[:2] == svs[:2]:
            p = self.config.Read("Perspective")
        #print "Config: " + str(p)
        if p:
            self._mgr.LoadPerspective(p)
            for menuid in self.panesByMenu:
                self.menuBar.Check(menuid, self._mgr.GetPane(self.panesByMenu[menuid]).IsShown())
        else:
            self._mgr.Update()
        if len(sys.argv) > 1:
            self.OnDropFiles([sys.argv[1]])
        self.frame.Show()



###############################################################
# Name:             Tokenizer.py
# Purpose:     Abstract tokenizer
# Author:         Luca Allulli (webmaster@roma21.it)
# Created:     2009-01-31
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:     GNU GPL v2
##############################################################

import re


class Token(object):
    def __init__(self, token, start, end, content):
        object.__init__(self)
        self.token = token
        self.start = start
        self.end = end
        self.content = content

    def __str__(self):
        return '%s, %d, %d, %s' % (str(self.token), self.start, self.end, self.content)


class TokenType(object):
    def __init__(self, r, name):
        object.__init__(self)
        self.r = re.compile(r)
        self.name = name

    def __str__(self):
        return self.name


class Tokenizer(object):
    def __init__(self, l):
        object.__init__(self)
        self.line = l
        self.pos = 0
        self.state = self.normal
        self.prevToken = None
        self.repeatToken = False

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.repeatToken:
            self.repeatToken = False
            return self.prevToken
        if self.pos >= len(self.line):
            #End of line
            raise StopIteration
        for t in self.transition[self.state]:
            # print("Trying: %s" % (t[0],))
            m = t[0].r.match(self.line, self.pos)
            if m != None:
                # print("Matched!")
                break
        if m == None:
            return None
        tok = Token(t[0], self.pos, m.end(0), m.group(0))
        ##print (type(m.group(0)), m.group(0))
        self.state = t[1]
        ##print("Found %s" % (tok,))
        self.pos = m.end(0)
        self.prevToken = tok
        return tok

    def Repeat(self, repeat = True):
        self.repeatToken = repeat



# -*- coding: utf-8 -*- 

import wx

_ = wx.GetTranslation

# -*- coding: utf-8 -*- 

###########################################################################
## Python code generated with wxFormBuilder (version Sep  8 2010)
## http://www.wxformbuilder.org/
##
## PLEASE DO "NOT" EDIT THIS FILE!
###########################################################################

###########################################################################
## Class UpdateDialog
###########################################################################

class UpdateDialog ( wx.Dialog ):
    
    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = _("New updates are available"), pos = wx.DefaultPosition, size = wx.Size( 258,250 ), style = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER )
        
        self.SetSizeHintsSz( wx.DefaultSize, wx.DefaultSize )
        
        bSizer20 = wx.BoxSizer( wx.VERTICAL )
        
        self.updatesSizer = wx.BoxSizer(wx.VERTICAL)
        
        bSizer20.Add( self.updatesSizer, 1, wx.ALL|wx.EXPAND, 5 )
        
        self.donate = wx.Button( self, wx.ID_ANY, _(">> Donate now! <<"), wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer20.Add( self.donate, 0, wx.ALL|wx.EXPAND, 5 )
        
        self.SetSizer( bSizer20 )
        self.Layout()
        
        self.Centre( wx.BOTH )
        
        # Connect Events
        self.donate.Bind( wx.EVT_BUTTON, self.OnDonate )
    
    def __del__( self ):
        pass
    
    
    # Virtual event handlers, overide them in your derived class
    def OnDonate( self, event ):
        event.Skip()
    

